import geopandas as gpd
from shapely.geometry import Polygon, Point, LineString, MultiPolygon
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import random
import math
import numpy as np

polygon_min_x = None  # The left most point of the entire polygon


# Read a json file with relation data and send to response object
def get_relation(self, body_of_water: str):  # NB: implement body_of_water
    divided_map = []  # List to store map shapes while splitting

    # NB: test polygon, remove after testing
    polygons = [circle_polygon()]

    # Divide all polygons into sections
    for polygon in polygons:

        cell_size = 0.4
        # Divide the length and with of polygon into a grid of equally sized parts
        grid_lines = create_grid_coords(polygon, cell_size)

        vrt_lines = grid_lines[0]  # Vertical grid coordinates
        hrz_lines = grid_lines[1]  # Horizontal grid coordinates

        # Cut polygon into horizontal sections, bottom to top
        for hrz_line in hrz_lines:

            # Split shape into upper and lower section as hrz_line as divider
            divided_poly = cut_polygon_by_points(polygon, hrz_line, cell_size)
            horizontal_section = divided_poly[0]  # Save upper horizontal section

            polygon = divided_poly[1]  # Set polygon to the remaining, un-split shape for next iteration

            if not horizontal_section or not divided_poly[0]:
                continue

                # Cut each horizontal section into vertical sections, right to left
            for vrt_line in vrt_lines:
                if len(horizontal_section.exterior.coords) < 3:
                    break  # Break from loop im remaining section has no coordinates

                # Split the horizontal section into two vertical parts
                vertical_parts = cut_polygon_by_points(horizontal_section, vrt_line, -0.1)

                divided_map.append(vertical_parts[0])  # Append split vertical sections to final list of shapes

                # Set horizontal_section to the remaining, un-split, horizontal section for next iteration
                horizontal_section = vertical_parts[1]

            polygon = divided_poly[1]  # Set polygon to the remaining, un-split shape for next iteration

        divided_map.append(polygon)

        break

    tiles = gpd.GeoDataFrame(geometry=divided_map)

    # NB test plot
    fig, ax = plt.subplots()
    ax.set_aspect(1)
    # ax.xaxis.set_major_locator(ticker.MultipleLocator(0.2))

    for i, polygon in enumerate(tiles['geometry']):
        random_color = "#{:06x}".format(random.randint(0, 0xFFFFFF))

        x, y = polygon.exterior.xy
        ax.plot(x, y, color='blue', alpha=0.5)
        ax.fill(x, y, color=random_color, alpha=1)

    plt.show()

    # Convert GeoDataFrame to GeoJSON
    tiles_json = tiles.to_json()

    # Set headers
    self.send_response(200)
    self.send_header("Content-type", "application/json")
    self.end_headers()

    # Write GeoJSON to response object
    self.wfile.write(tiles_json.encode('utf-8'))


# Takes a polygon and divides its coordinates into two shapes, where divisor is a
# coordinate that defines the point of division.
def cut_polygon_by_points(polygon: Polygon, divisor: float, cell_size: float):
    # Extract polygon exterior coordinates
    exterior_coords = list(polygon.exterior.coords)

    # Initialize lists to store coordinates of new shapes after split
    split_shape = []
    remaining_shape = []

    # Loop through points and check which side of the division line they are
    if cell_size > 0:  # Horizontal split
        for point in exterior_coords:
            point = Point(point)  # Convert coordinates to Shapely Point object
            if point.y < divisor:  # Check if point is over or below divisor
                split_shape.append(point)  # Append to appropriate shape
            else:
                remaining_shape.append(point)

        # Populate newly created edges with points to facilitate vertical cutting
        populated_shapes = populate_new_edge(split_shape, remaining_shape, cell_size, divisor)
        if populated_shapes is not None:
            split_shape = populated_shapes[0]
            remaining_shape = populated_shapes[1]

    else:  # Vertical split
        for point in exterior_coords:
            point = Point(point)  # Convert coordinates to Shapely Point object
            if point.x < divisor:
                split_shape.append(point)
            else:
                remaining_shape.append(point)

    # Check if the split_shape has enough coordinates to create a polygon
    if len(split_shape) < 3:
        # print("Not enough coordinates to create valid polygon: Split shape: ", len(split_shape))
        split_shape = None
    else:
        split_shape.append(split_shape[0])  # Append first coord to create closed loop

    # Check if the remaining_shape has enough coordinates to create a polygon
    if len(remaining_shape) < 3:
        # print("Not enough coordinates to create valid polygon: Remaining shape: ", len(remaining_shape))
        remaining_shape = None
    else:
        remaining_shape.append(remaining_shape[0])  # Append first coord to create closed loop

    # Return split polygons as Shapely Polygon objects
    return Polygon(split_shape), Polygon(remaining_shape)


# Generate grid of equally spaced x and y coordinates where the grid size is determined by cell_size
def create_grid_coords(polygon: Polygon, cell_size: float):
    # Define boundaries of grid
    min_x, min_y, max_x, max_y = polygon.bounds

    global polygon_min_x
    polygon_min_x = min_x

    # Divide grid into sections of size *cell_size
    x_step = max((max_x - min_x) / 10, cell_size)
    y_step = max((max_y - min_y) / 10, cell_size)

    x_coords = np.arange(min_x, max_x + x_step, x_step)
    y_coords = np.arange(min_y, max_y + y_step, y_step)

    return x_coords, y_coords


# NB: only for testing
def circle_polygon():
    circle_points = []
    num_points = 500
    center_x, center_y = 0.0, 0.0
    radius = 6

    for i in range(num_points):
        angle = 2 * np.pi * i / num_points
        x = center_x + radius * np.cos(angle)
        y = center_y + radius * np.sin(angle)
        circle_points.append((x, y))
    circle_points.append(circle_points[0])  # Close the circle

    return Polygon(circle_points)


# Adds equally spaced points along an edge that is created after a polygon is cut.
def populate_new_edge(split_shape, remaining_shape, cell_size: float, divisor: float):
    # Prepend new points onto the newly created edge to facilitate vertical splitting
    if split_shape is not None and polygon_min_x is not None:
        # Define starting point with an x-value that will be common for all polygons
        starting_point = polygon_min_x

        print("Hit main if")
        # Create list of corners
        corners = []

        divisor_range = ((divisor - 0.1),(divisor + 0.1)) # Define tolerance

        # Find all corners of split shape. Corner = point that intersects divisor
        for point in split_shape:
            if divisor_range[0] < point.y < divisor_range[1]:
                corners.append(point)
                print("Hit point in split_shape")

        if not corners or len(corners) < 2:
            print("no corners :(")
            return None

        # Sort corners in ascending order (left to right) based on x coordinate
        sorted_corners = sorted(corners, key=lambda p: p.x)

        while starting_point < sorted_corners[0].x:  # Increment starting point until it is withing the polygons bounds
            starting_point += cell_size

        # if starting_point < left_corner.x: # NB: optimised substitute for previous while loop, requires testing
        #    starting_point += cell_size * math.floor(starting_point - left_corner.x)

        # Insert new points with cell_size spacing while starting_point is within bounds
        while starting_point < sorted_corners[-1].x:
            split_shape.insert(0, (starting_point, divisor))  # NB may have to add/subtract small offset of 0.00001
            remaining_shape.insert(-1, (starting_point, divisor))  # Prepend new point to shape
            print("Hit point insertion")

            starting_point += cell_size

        return split_shape, remaining_shape


